from jaepeto.structure.classes import FileAnalyser
from jaepeto.structure.entrypoint import (
    get_script_entrypoints,
    group_imports_at_level,
    infer_important_functions,
    resolve_import_paths,
)
from jaepeto.structure.hello_world import DevOnboardingParser, OnboardingParser



"""
Third-party tools
e.g., Coverage, Docker
"""
import os
import re
from pathlib import Path
from typing import List, Optional, Tuple

import requests

from jaepeto.utils import post


class ContainerParser:
    def __init__(self) -> None:
        self.file: Optional[str] = None

    def parse(self, project_dir: Path) -> Optional[str]:
        for spelling in ("dockerfile", "Dockerfile", "DOCKERFILE"):
            if spelling not in os.listdir(project_dir):
                continue

            self.file = spelling

            with open(project_dir / spelling, "r") as file:
                docker_contents = file.readlines()

            return post("parse-container", docker_contents)
        return None


class CIParser:
    def __init__(self) -> None:
        self.ci_names: List[str] = []
        self.ci_summaries: List[Tuple[str, str]] = []

    def parse(self, project_dir: Path) -> None:
        """
        Look through project for presence of CI files.

        Detects CI files and parses the files
        to infer some information about what the script is doing.
        """
        ci_files = {
            "Travis": [".travis.yml", "travis.yml", ".travis.yaml", "travis.yaml"],
            "GitHub Actions": [f".github{os.sep}workflows"],
            "Circle CI": [f".circleci{os.sep}config.yml"],
            "GitLab CI": [".gitlab-ci.yml"],
        }

        for ci_provider, file_names in ci_files.items():
            for file_name in file_names:
                if (project_dir / file_name).exists():
                    self.ci_names.append(ci_provider)
                    self.parse_ci_files(ci_provider, project_dir / file_name)
                    break

    def parse_ci_files(self, ci_provider: str, file_path: Path) -> None:
        """
        Parse a CI file for useful information.

        Information includes when the scripts are executed
        and what they do (i.e. what type of testing, or to where something is deployed).
        If a CI file is parsed, information is added to an attribute of this class.
        Only parses GitHub actions currently.

        Parameters
        ----------
        ci_provider : str
            The name of the CI provider detected (e.g. "GitLab CI")
        file_path : pathlib.Path
            Path to the file detected which signals presence of CI
        """
        if ci_provider == "GitHub Actions":
            for github_action in file_path.glob("*.yml"):
                github_action_name = github_action.stem

                with open(github_action, "r") as f:
                    contents = f.read()

                ci_summary = post("parse-ci/github", contents)
                self.ci_summaries.append((github_action_name, ci_summary))

    def __str__(self) -> str:
        if self.ci_names:
            if len(self.ci_names) == 1:
                arch = f"The project uses {self.ci_names[0]} for CI/CD."
            else:
                arch = f"The project uses multiple CI/CD providers: {', '.join(self.ci_names)}"

            # Table of CI file purposes, if present
            if self.ci_summaries:
                arch += "\n\n| CI File | Purpose |\n|:----|:----|"

            for ci_file, ci_summary in self.ci_summaries:
                arch += f"\n| {ci_file} | {ci_summary if ci_summary else ''} |"
        else:
            arch = "No CI/CD config files were detected."

        return arch



"""
Detect important environmental packages/libraries.
"""
import os
import re
from pathlib import Path
from typing import List

from jaepeto.utils import post


def read_packages(project_dir: Path) -> List[str]:
    packages = []

    if "requirements.txt" in os.listdir(project_dir):
        with open(project_dir / "requirements.txt", "r") as f:
            contents = f.readlines()

        if new_packages := post("parse-reqs/requirements", contents):
            packages.extend(new_packages)

    for _file in ("environment.yml", "environment.yaml"):
        if _file in os.listdir(project_dir):
            with open(project_dir / _file, "r") as f:
                contents = f.readlines()

            if new_packages := post("parse-reqs/environment", contents):
                packages.extend(new_packages)

    return packages



"""
Generate changelog.
"""
import re
from pathlib import Path
from typing import Optional

from jaepeto.utils.git import get_commits, list_tags


def catalog_commit(commit: str):
    commit_types = {
        "fixed": "Fixed",
        "fixes": "Fixed",
        "fix": "Fixed",
        "bugfix": "Fixed",
        "solve": "Fixed",
        "solves": "Fixed",
        "solved": "Fixed",
        "close": "Fixed",
        "closes": "Fixed",
        "closed": "Fixed",
        "corrects": "Fixed",
        "correct": "Fixed",
        "corrected": "Fixed",
        "create": "Added",
        "creates": "Added",
        "created": "Added",
        "make": "Added",
        "makes": "Added",
        "made": "Added",
        "write": "Added",
        "wrote": "Added",
        "add": "Added",
        "adds": "Added",
        "added": "Added",
        "deprecate": "Changed",
        "deprecates": "Changed",
        "deprecated": "Changed",
        "list": "Changed",
        "lists": "Changed",
        "use": "Changed",
        "uses": "Changed",
        "fetch": "Changed",
        "fetches": "Changed",
        "fetched": "Changed",
        "log": "Changed",
        "logs": "Changed",
        "logged": "Changed",
        "improve": "Changed",
        "improves": "Changed",
        "improved": "Changed",
        "print": "Changed",
        "prints": "Changed",
        "printed": "Changed",
        "rewrite": "Changed",
        "rewrote": "Changed",
        "rewrit": "Changed",
        "refactor": "Changed",
        "change": "Changed",
        "changes": "Changed",
        "changed": "Changed",
        "move": "Changed",
        "moves": "Changed",
        "moved": "Changed",
        "update": "Changed",
        "updates": "Changed",
        "updated": "Changed",
        "tweak": "Changed",
        "tweaks": "Changed",
        "tweaked": "Changed",
        "remove": "Removed",
        "removes": "Removed",
        "removed": "Removed",
        "delete": "Removed",
        "deletes": "Removed",
        "deleted": "Removed",
        "deprecate": "Deprecated",
        "deprecates": "Deprecated",
        "deprecated": "Deprecated",
    }

    commit_lower = commit.lower()
    cleaned_commit = None
    start_word = commit_lower.split(" ")[0]

    try:
        commit_type = commit_types[start_word]
        cleaned_commit = commit_lower.replace(start_word, "").strip().capitalize()
        return commit_type, cleaned_commit
    except KeyError:
        pass

    return None, None


class ChangelogSection:
    def __init__(self, title: str) -> None:
        self.title = title
        self._commits = []

    def add_commit(self, commit_title: str):
        if not re.match(r"\s*[\*-]\s", commit_title):
            commit_title = "* " + commit_title

        self._commits.append(commit_title.rstrip())

    def has_content(self) -> bool:
        return len(self._commits) > 0

    def remove_duplicates(self) -> None:
        unique_commits = []
        for commit in self._commits:
            if commit not in unique_commits:
                unique_commits.append(commit)

        self._commits = unique_commits

    def __str__(self) -> str:
        content = f"### {self.title}\n\n"

        for _commit in self._commits:
            content += _commit + "\n"

        return content


class ChangelogRelease:
    def __init__(self, version: str, date: str) -> None:
        self._version = version
        self._date = date

        self._sections = {
            "Added": ChangelogSection("Added"),
            "Changed": ChangelogSection("Changed"),
            "Fixed": ChangelogSection("Fixed"),
            "Removed": ChangelogSection("Removed"),
            "Deprecated": ChangelogSection("Deprecated"),
        }

    def read(self, lines) -> None:
        for i, line in enumerate(lines):
            _line = line.lstrip("#").strip()

            if _line.lower() in ("added", "changed", "fixed", "removed", "deprecated"):
                title = _line.capitalize()

                for commit_line in lines[i + 1 :]:
                    if re.match(r"\s*[\*-]\s", commit_line):
                        self._sections[title].add_commit(commit_line)
                    elif re.match(r"#+\s", commit_line):
                        break

    def generate(self, start_date: str) -> None:
        commits = get_commits(since_date=start_date, until_date=self._date)

        for com in commits:
            if title := com["title"]:
                commit_type, clean_commit = catalog_commit(title)
                if commit_type is not None and self.validate_commit(
                    clean_commit, commit_type
                ):
                    self._sections[commit_type].add_commit(clean_commit)

        for section in self._sections.keys():
            self._sections[section].remove_duplicates()

    def validate_commit(self, commit: str, commit_type: str) -> bool:
        """Confirm if commit is useful for CHANGELOG."""
        commit_words = commit.strip().lower().split(" ")

        if len(commit_words) == 1:
            if "." in commit_words[0] or commit_words[0] in ("license", "changelog"):
                # Naively assume it's a file if has "."
                # TODO: improve this assumption
                # Files are okay for "Added" and "Removed" sections
                return commit_type in ("Added", "Removed")
            else:
                return False

        if len(commit_words) == 2:
            if "bug" in commit_words:
                # Too generic -> not valid
                return False

        if "i" in commit_words:
            # Wrong tone
            return False

        return True

    def header(self) -> str:
        return f"## {self._version} - {self._date}"

    def __str__(self) -> str:
        content = self.header() + "\n\n"

        for section in self._sections.values():
            if section.has_content():
                content += str(section) + "\n\n"

        return content


class ChangelogUnreleased(ChangelogRelease):
    def __init__(self) -> None:
        super().__init__("Unreleased", "2100-12-31")

    def header(self) -> str:
        return f"## {self._version}"


class ChangelogDoc:
    def __init__(self, changelog_path: Path) -> None:
        self._releases = {}
        self._changelog_path = changelog_path
        self._tags = list_tags()

    def _get_tag_date(self, tag_name: str) -> str:
        for tag, tag_date in self._tags:
            if tag == tag_name:
                return tag_date

    def read(self) -> None:
        if not self._changelog_path.exists():
            return

        with open(self._changelog_path, "r") as f:
            content = f.readlines()

        curr_version = None
        version_date = None
        curr_lines = []

        def _read_release():
            if curr_version is not None:
                if curr_version.lower() == "unreleased":
                    new_release = ChangelogUnreleased()
                else:
                    new_release = ChangelogRelease(curr_version, version_date)

                new_release.read(curr_lines)
                self._releases[curr_version] = new_release

        for line in content:
            _line = line.lstrip("#").strip()
            _line_parts = _line.split("-")
            _ver = _line_parts[0].strip().lstrip("v")

            if (
                re.match(r"[0-9]{1,2}\.[0-9]{1,2}\.[0-9]{1,2}", _ver)
                or _ver.lower() == "unreleased"
            ):
                _read_release()

                curr_version = _ver
                curr_lines = []
                if len(_line_parts) == 2:
                    version_date = _line_parts[1].strip()
                else:
                    version_date = str(self._get_tag_date("v" + _ver))

            else:
                curr_lines.append(line)

        _read_release()

    def generate(self):
        self.read()

        if "Unreleased" not in self._releases or (
            len(self._tags) > 0 and self._tags[0][0].lstrip("v") not in self._releases
        ):
            # Either no unreleased or existing unreleased has become new tagged version
            self._releases["Unreleased"] = ChangelogUnreleased()

        if len(self._tags) > 0:
            unreleased_start_date = self._tags[0][1]
        else:
            unreleased_start_date = "1999-01-01"

        self._releases["Unreleased"].generate(unreleased_start_date)

        for i, (tag_name, tag_date) in enumerate(self._tags):
            if tag_name.lstrip("v").strip() in self._releases:
                continue

            if i == len(self._tags) - 1:
                section_start = "1999-01-01"
            else:
                section_start = self._tags[i + 1][1]

            release = ChangelogRelease(tag_name, tag_date)
            release.generate(section_start)
            self._releases[tag_name.lstrip("v").strip()] = release

    def __str__(self) -> str:
        content = "# Changelog\n\nBased on KeepAChangelog.\n**Generated by Jaepeto.**"

        for _, release in self._releases.items():
            content += "\n\n" + str(release)

        return content

    def save(self) -> None:
        with open(self._changelog_path, "w") as f:
            f.write(str(self))



"""
Identify and analyse how classes are used in the project.
"""
import ast
import os
from collections import defaultdict
from functools import reduce
from pathlib import Path
from typing import DefaultDict, Dict, List, Optional, Tuple

from jaepeto.structure.entrypoint import identify_module_imports, resolve_import_paths
from jaepeto.utils import parse_all_type_hints


class FileAnalyser:
    def __init__(self, module_name: str, module_root: Path, max_depth: int = 3):
        self.__module_name = module_name
        self.__module_root = module_root
        self.__max_depth = max_depth

    def fetch_class_names(self, file_contents: str) -> List[str]:
        """
        Fetch all class names present in a file.

        If no class names are present, return an empty list.

        Parameters
        ----------
        file_contents : str
            Contents of a file as a single string

        Returns
        -------
        list of str
            List of all class names defined in the file
        """
        tree = ast.parse(file_contents)
        names: List[str] = []

        for node in tree.body:
            if isinstance(node, ast.ClassDef):
                names.append(node.name)

        return names

    def get_class_names(self) -> List[Tuple[str, str]]:
        """
        Fetch the names of all custom classes created in a project.

        Returns
        -------
        list of (str, str) tuples
            First element of tuple is the class name
            Second element of tuple is the "." separated import path to the class name, excluding the class
                e.g. ("MyClass", "root.subdir.subsubdir.file")
        """
        class_names: List[str] = []

        for py_file in self.__module_root.rglob("*.py"):
            file_prefix = ".".join(
                str(py_file.relative_to(self.__module_root).with_suffix("")).split(
                    os.sep
                )
            )

            if len(file_prefix.split(".")) > self.__max_depth:
                continue

            with open(py_file, "r") as f:
                names = self.fetch_class_names(f.read())

            class_names.extend([(name, file_prefix) for name in names])
        return class_names

    def get_base_classes(self, class_path: str) -> Optional[List[str]]:
        """
        Fetch the names of all base classes inherited by a class.

        Parameters
        ----------
        class_path : str
            Path of class seperated by ".". eg root.subdir.subsubdir.file
        src_path: Path
            root path of project

        Returns
        -------
        None if there are no base classes inheritted by the class
        List of Base names as strings
        """
        class_name = class_path.split(".")[-1]
        class_absolute_path = self.__module_root / (
            os.sep.join(class_path.split(".")[1:-1]) + ".py"
        )
        with open(str(class_absolute_path), "r") as f:
            data = f.read()
            for node in ast.parse(data).body:
                if isinstance(node, ast.ClassDef) and node.name == class_name:
                    return (
                        None
                        if len(node.bases) == 0
                        else [parent.id for parent in node.bases]
                    )
            return None

    def parse(self) -> Dict[str, Optional[List[str]]]:
        """
        Fetch the names of all base classes inherited by a class.

        Returns
        -------
        Dictionary of important class as strings and the respective List of  string Base names
        as values
        """
        class_mapping: Dict = {}
        important_classes = self.resolve_class_uses()
        for important_class in important_classes:
            bases = self.get_base_classes(important_class)
            class_mapping[important_class] = (bases, important_classes[important_class])
        return class_mapping

    def resolve_class_uses(self) -> DefaultDict[str, list]:
        """
        Find every occurrence where a class is used as a type hint.

        Returns
        -------
        dict of str: list of str
            Keys are import paths to a class, e.g. `module.sub.file.ClassName`
            Values are list of path to every function where class is used as a type hint
        """

        class_uses = defaultdict(list)

        for py_file in self.__module_root.rglob("*.py"):
            local_path = ".".join(
                str(py_file.relative_to(self.__module_root).with_suffix("")).split(
                    os.sep
                )
            )
            if len(local_path.split(".")) > self.__max_depth:
                continue

            local_path = self.__module_name + "." + local_path
            local_module = ".".join(local_path.split(".")[:-1])  # exclude file

            with open(py_file, "r") as f:
                contents = f.read()

            imports = identify_module_imports(local_module, contents.split("\n"))
            type_hints = parse_all_type_hints(py_file)
            local_classes = self.fetch_class_names(contents)
            for func, types in type_hints.items():
                for import_path, import_statement in imports:
                    if import_statement in types:
                        class_uses[import_path].append((local_path + "." + func))

                for _type in types:
                    if _type in local_classes:
                        class_uses[local_path + "." + _type].append(
                            local_path + "." + func
                        )

        return class_uses



from argparse import ArgumentParser
from pathlib import Path

from jaepeto.docs.changelog import ChangelogDoc
from jaepeto.docs.technical_docs import (
    append_to_section,
    create_technical_doc,
    update_doc_code_sections,
)
from jaepeto.utils import parse_config, post


def generate_documentation() -> None:
    """
    Generate and save documentation.

    This is a CLI entrypoint to the codebase.
    The current working directory must be the root of the project
    for which documentation is to be generated.
    The root project must contain a `.jaepeto.ini` file.
    """
    parser = ArgumentParser("CLI entrypoint for documentation generation")
    parser.add_argument("document", nargs="?", default="technical")
    parser.set_defaults(changelog=False)
    args = parser.parse_args()

    config_path = Path.cwd()
    config = parse_config(config_path)
    arg = args.document.lower()

    if arg == "changelog":
        changelog_path = config_path / config["docs"]["changelog"]
        changelog = ChangelogDoc(changelog_path)
        changelog.generate()
        changelog.save()

        post("notify", None)
        print(f"Successfully generated CHANGELOG at {changelog_path}")
    elif arg == "code":
        tech_doc_path = (
            config_path / config["doc_dir"] / config["docs"]["technical_name"]
        ).with_suffix(".md")

        post("notify", None)

        update_doc_code_sections(tech_doc_path)
        print(f"Updated code snippets in {tech_doc_path}")
    else:
        tech_doc_path = (
            config_path / config["doc_dir"] / config["docs"]["technical_name"]
        ).with_suffix(".md")

        post("notify", None)

        create_technical_doc(
            config["src"],
            config["package"],
            config_path,
            tech_doc_path,
            config,
            project_title=config["name"],
            project_summary=config["description"],
        )
        print(f"Successfully generated technical documentation at {tech_doc_path}")



from pathlib import Path

from jaepeto.utils import post


def detect_requests_calls(file_contents: str) -> bool:
    available_request_methods = [
        method
        for method in ["get", "post", "put", "patch", "delete"]
        if f"requests.{method}(" in file_contents
    ]
    return len(available_request_methods) > 0


def detect_boto3_calls(file_contents: str) -> bool:
    return "boto3.resource" in file_contents


def detect_mongodb_calls(file_contents: str) -> bool:
    return "MongoClient" in file_contents


def detect_redis_calls(file_contents: str) -> bool:
    return "Redis(" in file_contents


def detect_gcp_calls(file_contents: str) -> bool:
    return "google.cloud" in file_contents or "from gcloud import " in file_contents


def describe_file_architecture(file_path: Path) -> str:
    with open(file_path, "r") as f:
        full_contents = f.read()

    checks = [
        detect_mongodb_calls,
        detect_boto3_calls,
        detect_requests_calls,
        detect_redis_calls,
        detect_gcp_calls,
    ]

    for arch_notifier in checks:
        if arch_notifier(full_contents):
            with open(file_path, "r") as f:
                contents = f.readlines()

            contents = [c.strip().strip("\n") for c in contents]
            return post("arch", contents)



import ast
import re
from pathlib import Path
from typing import List, Optional, Tuple, Union


def find_func_lines(node: Union[ast.FunctionDef, ast.ClassDef]):
    return node.lineno - 1, node.end_lineno - 1


def find_max_func_line(node: ast.FunctionDef):
    max_line = node.lineno

    if hasattr(node, "body"):
        for child in node.body:
            child_max = find_max_func_line(child)
            max_line = max(max_line, child_max)

    if isinstance(node, ast.If):
        for child in node.orelse:
            child_max = find_max_func_line(child)
            max_line = max(max_line, child_max)

    return max_line


def get_specific_python_function(function_name: str, file, max_length: int = 100):
    with open(file, "r") as source:
        tree = ast.parse(source.read())

    for node in tree.body:
        if not (isinstance(node, ast.FunctionDef) or isinstance(node, ast.ClassDef)):
            continue

        if node.name == function_name:
            func_start, func_end = find_func_lines(node)
            py_string = _read_python_func(file, func_start, func_end, max_length)

            # This is a hacky way to try to get working code. Not robust
            # TODO: improve and make robust
            try:
                ast.parse(py_string)
            except Exception:
                return _read_python_func(file, func_start, func_end + 1, max_length)
            else:
                return py_string


def get_func_lines(file):
    with open(file, "r") as source:
        tree = ast.parse(source.read())

    for node in tree.body:
        if isinstance(node, ast.FunctionDef):
            yield find_func_lines(node)


def get_funcs_as_string(
    file, min_length: int = 6, max_length: int = 100, include_docstring: bool = False
):
    for (func_start, func_end) in get_func_lines(file):
        if 1 + func_end - func_start >= min_length:
            func_string = _read_python_func(file, func_start, func_end, max_length)

            if not include_docstring:
                func_string = _remove_docstring(func_string)

            yield func_string


def _remove_docstring(func_string):
    """
    Remove docstring from a function call.

    Notes
    -----
    * Naively searches for triple quotes
    """
    split_func = func_string.split('"""')

    # TODO: Will incorrectly parse code if triple quotes present in code body
    if len(split_func) >= 3:
        split_func[0] = (
            split_func[0].rstrip() + "\n"
        )  # rstrip also removes \n. TODO: use raw strings to avoid this
        split_func[2] = split_func[2].lstrip("\n")

        del split_func[1]

    return "".join(split_func)


def _read_python_func(
    file: Path, min_line: int, max_line: int, max_line_difference: int
) -> str:
    """
    Read a subset of lines in a Python function.

    Parameters
    ----------
    min_line : int
        The starting line number (inclusive) which to read
    max_line : int
        The end line number (inclusive) which to read
    file : pathlib.Path
        The file to read
    max_line_different : int
        Maximum amount of lines to read. Overrides `max_line`.

    Returns
    -------
    str
        Subset of lines from the given file
    """
    with open(file, "r") as source:
        code = source.readlines()

    max_line = min(max_line, min_line + max_line_difference - 1)

    return "".join(code[min_line : max_line + 1])


def identify_main_functions(file: Path):
    """
    Extract function names called in script's __main__ entrypoint.

    If script does not have `if __name__ == "__main__"`,
    then gracefully return an empty list.
    For each call, only detects than an object has been called;
    it does not check whether it's a method, a function, or something else.

    Parameters
    ----------
    file : pathlib.Path
        The path to the python file in which to search for the entrypoint

    Returns
    -------
    bool
        Whether the file has a __main__ entrypoint
    list of str
        The name of any object called with a __main__ entrypoint
    """
    with open(file, "r") as f:
        tree = ast.parse(f.read())

    called_funcs: List[str] = []
    has_main_entrypoint = False

    for _node in tree.body:
        if not isinstance(_node, ast.If):
            continue

        _node_test = _node.test

        if not isinstance(_node_test, ast.Compare):
            continue

        test_left = _node_test.left

        if not isinstance(test_left, ast.Name):
            # Might be a subscripted comparator, for example
            # Whatever it is, it's certainly not __name__
            continue

        if (
            test_left.id == "__name__"
            and isinstance(_node_test.ops[0], ast.Eq)
            and _node_test.comparators[0].value == "__main__"
        ):
            has_main_entrypoint = True

            for sub_node in _node.body:
                if isinstance(sub_node, ast.Expr) or isinstance(sub_node, ast.Assign):
                    sub_node_value = sub_node.value

                    if isinstance(sub_node_value, ast.Call):
                        called_funcs.extend(parse_ast_call(sub_node_value))

    return has_main_entrypoint, called_funcs


def parse_ast_call(node: ast.Call) -> List[str]:
    called_funcs: List[str] = []

    if _func_call := _parse_ast_func(node.func):
        called_funcs.append(_func_call)

    for _func_arg in node.args:
        if isinstance(_func_arg, ast.Call):
            called_funcs.extend(parse_ast_call(_func_arg))

    for _func_kwarg in node.keywords:
        if isinstance(_func_kwarg.value, ast.Call):
            called_funcs.extend(parse_ast_call(_func_kwarg.value))

    return called_funcs


def _parse_ast_func(node) -> Optional[str]:
    if isinstance(node, ast.Attribute):
        return node.attr
    elif isinstance(node, ast.Name):
        return node.id


def parse_all_type_hints(file: Path):
    """
    Extract type hints for each function/method in a given file.
    """
    with open(file, "r") as f:
        tree = ast.parse(f.read())

    function_hints = {}

    for node in tree.body:
        if isinstance(node, ast.FunctionDef):
            function_hints[node.name] = parse_function_type_hint(node)

    return function_hints


def parse_function_type_hint(code: ast.FunctionDef) -> List[str]:
    type_hints = []

    for arg in code.args.args:
        type_hints.extend(parse_type_hint(arg.annotation))

    return list(set(type_hints))


def parse_type_hint(annotation) -> List[str]:
    """
    Convert an ast annotation into a readable type.

    High-level types like "List" and "Tuple" are not resolved further,
    e.g. not to "list of list of str".
    Function can parse 3.10-style | operator.

    Parameters
    ----------
    annotation : ast object
        The ast annotation object to parse

    Returns
    -------
    list of str
        All high-level types conferred by the annotation
    """
    if isinstance(annotation, ast.Name):
        types = annotation.id.split("|")
        return [_type.strip() for _type in types]
    elif isinstance(annotation, ast.Attribute):
        return [parse_type_hint(annotation.value)[0] + "." + annotation.attr]
    elif isinstance(annotation, ast.Subscript):
        if annotation.value.id == "Union":
            types = []
            annotation_slice = annotation.slice

            # Slight API change between python 3.8 and 3.9
            if hasattr(annotation_slice, "elts"):
                # Python 3.9+ API
                for sub_type in annotation_slice.elts:
                    types.extend(parse_type_hint(sub_type))
            elif hasattr(annotation_slice, "value") and hasattr(
                annotation_slice.value, "elts"
            ):
                for sub_type in annotation_slice.value.elts:
                    types.extend(parse_type_hint(sub_type))

            return types
        elif annotation.value.id == "Optional":
            # TODO: get the subscripted type
            return []
        else:
            # Only care about list/tuple at high level atm.
            # Don't need to get exact type
            return [annotation.value.id.lower()]
    elif isinstance(annotation, ast.Constant):
        return [annotation.value]
    else:
        return []


def parse_argument_parser(file_contents):
    tree = ast.parse(file_contents)
    identifier = get_parser_identifier(tree)
    arguments = []

    for node in ast.walk(tree):
        if isinstance(node, (ast.Call)):
            func = node.func
            if (
                isinstance(func, ast.Attribute)
                and func.attr == "add_argument"
                and func.value.id == identifier
            ):
                arg = node.args[0].value
                arguments.append(arg)
    return arguments


def get_parser_identifier(tree):
    for node in ast.walk(tree):

        if isinstance(node, (ast.Assign)):
            value = node.value
            if isinstance(value, (ast.Call)):
                call = value.func
                if isinstance(call, (ast.Attribute)) and call.attr == "ArgumentParser":
                    name = node.targets[0].id
                    return name



"""
Parse user config
"""
import os
from configparser import ConfigParser, NoOptionError
from pathlib import Path
from typing import Dict, Optional, Union

import toml
from dotenv import load_dotenv


def _clean_config_dir(value: str) -> str:
    """
    Remove errant / and . from directory values in config.

    Leading `./` and trailing `/`
    are to be removed as the code does not expect paths
    in this format and consequently cannot resolve the paths.

    Parameters
    ----------
    value : str
        The unclean directory

    Returns
    -------
    str
        A clean directory
    """
    value = value.lstrip("./")
    value = value.rstrip("/")
    return value


def parse_jaepeto_ini(config_file: Union[Path, str]) -> Dict:
    config = ConfigParser()
    config.read(config_file)

    project_name = config.get("project", "name")
    project_name = project_name if project_name else "Project"

    src_dir = _clean_config_dir(config.get("project", "srcdir"))
    try:
        package = config.get("project", "package")
    except NoOptionError:
        package = src_dir if "src/" not in src_dir else src_dir.split("src/")[1]

    try:
        description = config.get("project", "description")
    except NoOptionError:
        description = "<SHORT PROJECT DESCRIPTION>"

    project_info = {
        "name": project_name,
        "description": description,
        "src": _clean_config_dir(config.get("project", "srcdir")),
        "doc_dir": _clean_config_dir(config.get("docs", "docdir")),
        "package": package,
    }

    try:
        local = config.getboolean("project", "local")
    except NoOptionError:
        local = False

    try:
        api_key = config.get("project", "api_key")
    except NoOptionError:
        if not os.getenv("JAEPETO_API_KEY"):
            if local:
                print("Using local server.")
            else:
                raise RuntimeError(
                    "No API key detected. Contact `info AT documatic DOT com` to request one."
                )
    else:
        print(
            "Reading api key from config file is DEPRECATED. Place key in `.env` file. See README for more info."
        )
        os.environ["JAEPETO_API_KEY"] = config.get("project", "api_key")

    os.environ["JAEPETO_PROJECT_NAME"] = project_name
    os.environ["JAEPETO_RUN_LOCALLY"] = str(local)

    try:
        language = config.get("project", "language")
    except NoOptionError:
        language = "NoLanguage"

    if language.lower().strip().rstrip("3") != "python":
        print(
            f"""Jaepeto currently only works
        on Python (3), not {language}"""
        )

    # Documentation types
    project_info["docs"] = {
        "technical": True,
        "technical_name": "technical_doc",
        "changelog": "CHANGELOG.md",
    }

    # Documentation settings
    default_settings = {
        "num_class_uses_threshold": 3,
        "max_num_entrypoints_threshold": 10,
    }

    if config.has_section("settings"):
        settings = {}

        for settings_key, default_value in default_settings.items():
            try:
                settings[settings_key] = int(config.get("settings", settings_key))
            except (NoOptionError, ValueError):
                settings[settings_key] = default_value

        project_info["settings"] = settings
    else:
        project_info["settings"] = default_settings

    # Issues
    if config.has_section("issues"):
        issue_service = config.get("issues", "integration")
        if issue_service and issue_service.lower() != "github":
            project_info["issues"] = {}
        else:
            try:
                username = config.get("issues", "username")
                repo = config.get("issues", "repo")
            except:
                project_info["issues"] = {}
            else:
                project_info["issues"] = {"username": username, "repo": repo}
    else:
        project_info["issues"] = {}

    return project_info


def parse_pyproject(config_path: Path) -> Optional[Dict]:
    with open(config_path, "r") as f:
        config_str = f.read()
        config = toml.loads(config_str)

    if "tool" not in config or "jaepeto" not in config["tool"]:
        return None

    jaepeto_config = config["tool"]["jaepeto"]
    project_name = jaepeto_config.get("name")

    if not project_name:
        raise RuntimeError("No 'name' field (project name) in pyproject.toml")

    description = jaepeto_config.get("description", "<SHORT PROJECT DESCRIPTION>")

    srcdir = jaepeto_config.get("srcdir")

    if not srcdir:
        raise RuntimeError("No `srcdir` field (path to source code) in pyproject.toml")

    srcdir = _clean_config_dir(srcdir)

    package = jaepeto_config.get("package", srcdir)
    local = jaepeto_config.get("local", False)
    docdir = _clean_config_dir(jaepeto_config.get("docdir", "."))
    doc_name = jaepeto_config.get("docname", "technical_doc.md")

    if not doc_name.endswith(".md"):
        doc_name += ".md"

    os.environ["JAEPETO_PROJECT_NAME"] = project_name
    os.environ["JAEPETO_RUN_LOCALLY"] = str(local)

    if not os.getenv("JAEPETO_API_KEY"):
        print("No API key detected. Contact `info AT documatic DOT com` to request one")

    project_info = {
        "name": project_name,
        "description": description,
        "src": srcdir,
        "doc_dir": docdir,
        "package": package,
    }

    # Documentation types
    project_info["docs"] = {
        "technical": True,
        "technical_name": doc_name,
        "changelog": "CHANGELOG.md",
    }

    # Documentation settings
    project_info["settings"] = {
        "num_class_uses_threshold": 3,
        "max_num_entrypoints_threshold": 10,
    }

    return project_info


def parse_config(project_root: Path) -> Dict:
    load_dotenv()

    if (ini_path := (project_root / ".jaepeto.ini")).exists():
        return parse_jaepeto_ini(ini_path)
    elif (pyproject_path := project_root / "pyproject.toml").exists():
        project_dir = parse_pyproject(pyproject_path)

        if project_dir:
            return project_dir
        else:
            raise RuntimeError("Invalid config for [tool.jaepeto] in pyproject.toml")
    else:
        raise RuntimeError("No config file detected")


def create_config():
    """
    Create a minimal config file.

    Must be called when project root is current working directory.
    Some information must be filled out by a user, such as `api_key`.
    `srcdir` is assumed to be `src` or just home directory.

    Notes
    -----
    * Writes config file to CWD / ".jaepeto.ini"
    """
    project = Path.cwd()
    config_path = project / ".jaepeto.ini"

    if config_path.exists():
        print(".jaepeto.ini already exists!")
        return

    config = ConfigParser()

    project_name = project.stem

    src_dir = project / "src"
    if src_dir.is_dir():
        srcdir = "src"
        for _file in src_dir.glob("*"):
            if _file.is_dir():
                package = _file.stem
                break
    else:
        srcdir = "."
        package = "."

    config.add_section("project")
    config.set("project", "name", project_name)
    config.set("project", "description", "<ENTER SHORT PROJECT DESCRIPTION>")
    config.set("project", "language", "python")
    config.set("project", "api_key", "<ADD API KEY HERE>")
    config.set("project", "srcdir", srcdir)
    config.set("project", "package", package)

    config.add_section("settings")
    config.set("settings", "num_class_uses_threshold", "3")
    config.set("settings", "max_num_entrypoints_threshold", "10")

    config.add_section("docs")
    config.set("docs", "docdir", "./")
    config.set("docs", "changelog", "CHANGELOG")

    print(f"Creating config file at {config_path}")
    with open(config_path, "w") as f:
        config.write(f)



import re
import subprocess
from collections import defaultdict
from datetime import datetime
from typing import Optional, Tuple

from jaepeto.utils.parse_python import get_specific_python_function

leading_4_spaces = re.compile("^    ")


def get_commits(since_date: Optional[str] = None, until_date: Optional[str] = None):
    if since_date and until_date:
        command = ["git", "log", "--stat", "--since", since_date, "--until", until_date]
    else:
        command = ["git", "log", "--stat"]

    lines = (
        subprocess.check_output(command, stderr=subprocess.STDOUT)
        .decode("utf-8")
        .split("\n")
    )

    commits = []
    current_commit = {}

    def save_current_commit():
        title = current_commit["message"][0]
        message = current_commit["message"][1:]
        changed_files = []

        if message and message[0] == "":
            del message[0]

        current_commit["title"] = title
        current_commit["message"] = "\n".join(message)
        current_commit["files"] = [
            _m.split("|")[0].strip() for _m in message if "|" in _m
        ]
        commits.append(current_commit)

    for line in lines:
        if not line.startswith(" "):
            if line.startswith("commit "):
                if current_commit:
                    save_current_commit()
                    current_commit = {}
                current_commit["hash"] = line.split("commit ")[1]
            else:
                try:
                    key, value = line.split(":", 1)
                    current_commit[key.lower()] = value.strip()
                except ValueError:
                    pass
        else:
            current_commit.setdefault("message", []).append(
                leading_4_spaces.sub("", line)
            )

    if current_commit:
        save_current_commit()

    commits.reverse()  # earliest commit first
    return commits


def find_earliest_commit_with_file(file: str, commits):
    # Assume commits are ordered earliest to latest
    for commit in commits:
        for commit_file in commit["files"]:
            if file in commit_file.lower():
                # Accepts partial match e.g. .travis -> .travis.yml
                return convert_datetime_to_date_string(commit["date"])

    raise RuntimeError(f"No commits involve file {file}")


def get_diff_for_file(file: str, commits):
    changes = []

    for i in range(1, len(commits)):
        main_commit = commits[i]

        if file not in main_commit["files"]:
            continue

        previous_commit_hash = commits[i - 1]["hash"]
        main_commit_hash = main_commit["hash"]

        lines = (
            subprocess.check_output(
                ["git", "diff", previous_commit_hash, main_commit_hash, file],
                stderr=subprocess.STDOUT,
            )
            .decode("utf-8")
            .split("\n")
        )

        lines = [l for l in lines if re.match("[+-][a-zA-Z]", l)]

        formatted_date = convert_datetime_to_date_string(main_commit["date"])
        changes.append((main_commit_hash, formatted_date, lines))

    return changes


def convert_datetime_to_date_string(datetime_string):
    return datetime.strptime(datetime_string, "%a %b %d %H:%M:%S %Y %z").strftime(
        "%Y-%m-%d"
    )


def get_line_changes_for_file(file: str):
    lines = (
        subprocess.check_output(
            ["git", "log", "-p", '--since="4 weeks ago"', "--", file],
            stderr=subprocess.STDOUT,
        )
        .decode("utf-8")
        .split("\n")
    )

    commits = []
    current_commit = {}

    def _save_current_commit():
        title = current_commit["message"][0]
        message = current_commit["message"][1:]
        changed_files = []

        if message and message[0] == "":
            del message[0]

        current_commit["title"] = title
        commit_message = "\n".join(message)

        if "additions" in current_commit:
            lines_added = len(current_commit["additions"])
        else:
            lines_added = 0

        if "deletions" in current_commit:
            lines_removed = len(current_commit["deletions"])
        else:
            lines_removed = 0

        commits.append((lines_added, lines_removed, current_commit["author"]))

    for line in lines:
        if not line.startswith(" "):
            if line.startswith("commit "):
                if current_commit:
                    _save_current_commit()
                    current_commit = {}
                current_commit["hash"] = line.split("commit ")[1]
            else:
                if line == "+" or re.match(r"\+[^\+]", line):
                    current_commit.setdefault("additions", []).append(line)
                elif line == "-" or re.match(r"-[^-]", line):
                    current_commit.setdefault("deletions", []).append(line)
                else:
                    try:
                        key, value = line.split(":", 1)
                        current_commit[key.lower()] = value.strip()
                    except ValueError:
                        pass
        else:
            current_commit.setdefault("message", []).append(
                leading_4_spaces.sub("", line)
            )

    if current_commit:
        _save_current_commit()

    authors = {}
    for _added, _removed, _author in commits:
        authors.setdefault(_author, 0)
        authors[_author] += _added + _removed

    if authors:
        common_author = max(authors, key=authors.get)
    else:
        common_author = None

    return sum([c[0] for c in commits]), sum([c[1] for c in commits]), common_author


def _get_tag_date(tag_name: str) -> str:
    return (
        subprocess.check_output(
            ["git", "log", "-1", "--format=%ai", tag_name], stderr=subprocess.STDOUT
        )
        .decode("utf-8")
        .split(" ")[0]
    )


def list_tags():
    tags = (
        subprocess.check_output(["git", "tag", "-n"], stderr=subprocess.STDOUT)
        .decode("utf-8")
        .split("\n")
    )
    tags.reverse()

    tags = [t.split(" ")[0] for t in tags]
    tags = [(t, _get_tag_date(t)) for t in tags if t]
    return tags


def get_func_update(
    filepath: str, object_name: str, commit: str, is_class: bool = False
) -> Optional[Tuple[str, str, str]]:
    """Track the evolution of a function/class/method from a given commit.

    Parameters
    ----------
    filepath : str
        The file in which the object is present

    Returns
    -------
    (str, str, str, str) or None
        The latest commit with the object
        The latest name of the object
        The latest file in which the object presides
        The latest code for the object
        or None if object cannot be tracked.
    """
    last_change = (
        subprocess.check_output(
            [
                "git",
                "blame",
                filepath,
                "-L",
                ":" + object_name,
                "--reverse",
                commit,
            ],
            stderr=subprocess.STDOUT,
        )
        .decode("utf-8")
        .split("\n")
    )[0]

    last_commit_change = last_change.split(" ")[0].strip()
    last_file_change = last_change.split(" ")[1].strip()

    if last_file_change.startswith("("):
        last_file_change = filepath

    if not _get_next_commit(last_commit_change):
        # Is most recent commit, therefore cannot be other changes
        return (
            last_commit_change.lstrip("^"),
            object_name,
            last_file_change,
            get_specific_python_function(object_name, last_file_change),
        )

    """if new_file_info := _find_new_object_file(
        object_name, last_commit_change, is_class=is_class
    ):  # has func moved files?
        new_object_file, new_commit = new_file_info
        return get_func_update(
            new_object_file, object_name, new_commit, is_class=is_class
        )"""
    if new_object_info := _find_new_object_name(
        object_name, last_commit_change, is_class=is_class
    ):  # has func changed name?
        new_object_name, new_commit = new_object_info
        return get_func_update(filepath, new_object_name, new_commit, is_class=is_class)
    else:
        return


def _find_new_object_file(
    object_name, commit: str, is_class: bool
) -> Optional[Tuple[str, str]]:
    obj_def = "def" if not is_class else "class"

    if next_commit := _get_next_commit(commit):
        git_diff = subprocess.check_output(
            ["git", "diff", commit, next_commit], stderr=subprocess.STDOUT
        ).decode("utf-8")

        if not (
            f"-{obj_def} {object_name}(" in git_diff
            and f"+{obj_def} {object_name}(" in git_diff
        ):
            return

        if new_file_match := re.search(
            rf"\+\+\+ b/([^\s]*)\n.*?\n.*?\+{obj_def} {object_name}\(", git_diff
        ):
            return new_file_match.group(1), next_commit


def _find_new_object_name(
    object_name: str, commit: str, is_class: bool
) -> Optional[Tuple[str, str]]:
    obj_def = "def" if not is_class else "class"

    if next_commit := _get_next_commit(commit):
        git_diff = (
            subprocess.check_output(
                ["git", "diff", commit, next_commit], stderr=subprocess.STDOUT
            )
            .decode("utf-8")
            .split("\n")
        )

        for idx, diff in enumerate(git_diff):
            if (
                diff.startswith(f"-{obj_def} {object_name}(")
                and idx < len(git_diff) - 1
            ):
                new_obj = git_diff[idx + 1]

                if new_obj.startswith(f"+{obj_def} "):
                    new_obj_name = (
                        git_diff[idx + 1].lstrip(f"+{obj_def}").split("(")[0].strip()
                    )
                    return new_obj_name, next_commit


def _get_next_commit(commit: str) -> Optional[str]:
    commit = commit.lstrip("^")

    if commit == get_current_commit():
        return None

    commit_tree = (
        subprocess.check_output(
            ["git", "log", "--reverse", "--ancestry-path", f"{commit}^..HEAD"],
            stderr=subprocess.STDOUT,
        )
        .decode("utf-8")
        .split("\nAuthor")
    )[
        1
    ]  # 0th entry is `commit`, 1st entry is the commit after `commit`

    next_commits = commit_tree.split("commit ")

    if len(next_commits) == 1:
        return None
    else:
        return next_commits[-1].strip()[:7]


def get_current_commit() -> Optional[str]:
    """
    Get the shortform of the current git commit, if using git.

    Returns
    -------
    str
        Shortform (first 7 digits) of the latest git commit, if git is used and there are commits.
        Otherwise, return None.
    """
    try:
        output = subprocess.check_output(["git", "log", "-1", "--oneline"]).decode(
            "utf-8"
        )
        return output.split(" ")[0]
    except subprocess.CalledProcessError:
        return None



"""
Inferring how to launch a project.
"""
import ast
import os
import re
from configparser import ConfigParser, NoOptionError, NoSectionError
from pathlib import Path
from textwrap import dedent
from typing import List, Optional

import requests

from jaepeto.utils import post


class PyVersionDetector:
    def __init__(self):
        self._version_reason = []

        self.PY35 = 35
        self.PY36 = 36
        self.PY37 = 37
        self.PY38 = 38
        self.PY39 = 39
        self.PY310 = 310
        self._version = None
        self.possible_versions = (
            self.PY36,
            self.PY37,
            self.PY38,
            self.PY39,
            self.PY310,
        )

        subscript_type_pattern = r"\[\w+:?\s?\w+\]"
        type_hint_pattern = rf"\s?\w+(?:{subscript_type_pattern})?\s?"
        assignment_pattern = rf"\w+\s?(?::{type_hint_pattern})?=\s?"

        # 3.6
        self._f_string_36 = []
        self._f_string_pattern = re.compile(r"fr?['\"](.*)[\'\"]", re.IGNORECASE)

        # 3.8
        self._walrus = []
        self._walrus_pattern = re.compile(r"if\s\(?\w*\s?:=\s?.+\)?")
        self._f_string_38 = []

        # 3.9
        self._dict_union = []
        self._dict_union_pattern = re.compile(
            rf"^[^'\"]*?(?:(?:[,\(]\s?)|(?:{assignment_pattern}))\w+\s?\|\s?\w+"
        )

        # 3.10
        self._pipe_optional = []
        self._pipe_pattern = re.compile(
            rf"^[^'\"].*?\w+ ?:{type_hint_pattern}\|{type_hint_pattern}[,\)=]"
        )

        self._type_hint = []
        self._match_operator = []

    @property
    def version(self) -> str:
        _ver = str(self._version) if self._version else str(self.PY37)
        _ver = _ver[0] + "." + _ver[1:]
        return _ver

    def has_fstring(self, code, filename) -> bool:
        return False
        # TODO: fix this method
        tree = ast.parse(code, filename)
        for node in ast.walk(tree):
            if isinstance(node, ast.JoinedStr):
                return any(
                    [
                        isinstance(value, ast.Constant) and "=" in value.value
                        for value in node.values
                    ]
                )

    def review(self, filename: str, code_contents: List[str]) -> None:
        __version = self.PY35 if not self._version else self._version

        for considered_version in self.possible_versions:
            if considered_version <= __version:
                continue

            if considered_version == self.PY36:
                self._match_36(filename, code_contents)
            elif considered_version == self.PY38:
                self._match_38(filename, code_contents)
            elif considered_version == self.PY39:
                self._match_39(filename, code_contents)
            elif considered_version == self.PY310:
                self._match_310(filename, code_contents)

    def _match_36(self, filename: str, code_contents: List[str]) -> None:
        for _code in code_contents:

            _code = _code.strip()
            if pipe_dict := self._f_string_pattern.search(_code):
                if self.has_fstring(_code, filename) == False:
                    self._f_string_36.append(filename)

                    if not self._version or self._version < self.PY36:
                        self._version_reason = []
                        self._version = self.PY36
                    self._version_reason.append(("f-string 3.6", filename))

    def _match_38(self, filename: str, code_contents: List[str]) -> None:
        for _code in code_contents:
            _code = _code.strip()
            if walrus := self._walrus_pattern.search(_code):
                self._walrus.append(filename)

                if not self._version or self._version < self.PY38:
                    self._version_reason = []
                    self._version = self.PY38
                self._version_reason.append(("walrus", filename))
            elif f_string := self._f_string_pattern.search(_code):
                if self.has_fstring(_code, filename):
                    self._f_string_38.append(filename)

                    if not self._version or self._version < self.PY38:
                        self._version_reason = []
                        self._version = self.PY38
                    self._version_reason.append(("f-string 3.8", filename))

    def _match_39(self, filename: str, code_contents: List[str]) -> None:
        for _code in code_contents:
            _code = _code.strip()
            if pipe_dict := self._dict_union_pattern.search(_code):
                self._dict_union.append(filename)

                if not self._version or self._version < self.PY39:
                    self._version_reason = []
                    self._version = self.PY39
                self._version_reason.append(("pipe dictionary union", filename))

    def _match_310(self, filename: str, code_contents: List[str]) -> None:
        for _code in code_contents:
            _code = _code.strip()
            if pipe := self._pipe_pattern.search(_code):
                self._pipe_optional.append(filename)

                if not self._version or self._version < self.PY310:
                    self._version_reason = []
                    self._version = self.PY310
                self._version_reason.append(("pipe type hint", filename))

    def __str__(self):
        if not self._version:
            return dedent(
                """
                    No exact compatable Python version has been detected.
                    Versions below 3.8 are compatable.
                    """
            )
        else:
            ver_str = f"The codebase is compatible with Python {self.version} and above, because of {self._version_reason[0][0]} in {self._version_reason[0][1]}."
            return ver_str


class SetupParser:
    def __init__(self) -> None:
        self.is_package = False
        self.pypi_package = None
        self.cli = {}

    def parse(self, project_dir: Path, package_name: Optional[str]) -> None:
        if (setup_py := (project_dir / "setup.py")).exists():
            self.is_package = True
            self.parse_setup(setup_py, setup_type="py")

        if (setup_cfg := (project_dir / "setup.cfg")).exists():
            self.is_package = True
            self.parse_setup(setup_cfg, setup_type="cfg")

        self._check_pypi(package_name)

    def parse_setup(self, file: Path, /, *, setup_type: str) -> None:
        with open(file, "r") as f:
            contents = f.read()

        if response := post(f"setup/{setup_type}", contents):
            if isinstance(response, list):
                for _key, _value in response:
                    self.cli[_key] = _value

    def _check_pypi(self, package_name: Optional[str]) -> None:
        """
        Check pypi to infer if package is globally pip-installable.

        Uses `requests` to look for the package on pypi.
        If present on pypi, assumes that that is the correct package.
        **Note:** This may lead to a false positive if the package shares
        a name with a project already on pypi.

        Parameters
        ----------
        package_name : str or None
            The local pip-installable package name for the codebase, if codebase is installable.
            Otherwise None.

        Notes
        -----
        * `pip search` has been taken down. If it resurfaces, that would be a more applicable check.
        """
        if package_name:
            if (
                requests.get(f"https://pypi.org/project/{package_name}/").status_code
                == 200
            ):
                self.pypi_package = package_name

    def __str__(self) -> str:
        setup_string = ""

        if self.is_package:
            setup_string += dedent(
                """
                Run `pip install -e .` in top-level directory to install
                package in local directory.
                """
            )

        if self.pypi_package:
            setup_string += f"Install from pypi with `pip install {self.pypi_package}`."

        return setup_string

    def read_entrypoints(self) -> str:
        description = ""

        cli_desc = ""
        for _command, _func in self.cli.items():
            cli_desc += f"* `{_command}` runs `{_func}`\n"

        if cli_desc:
            cli_desc = (
                "The project has the following CLI commands:\n\n" + cli_desc + "\n"
            )
            cli_desc += "Run a CLI command in a terminal to execute."
            description += cli_desc

        return description


class OnboardingParser:
    def __init__(self) -> None:
        self._version = PyVersionDetector()
        self._setup = SetupParser()

        self._reqs = False

    def parse(
        self, source_dir: Path, project_dir: Path, package_name: Optional[str] = None
    ) -> None:
        """
        Search a project codebase for "Getting Started" information for users.

        Relevant information includes minimum Python version
        and installation instructions.

        Parameters
        ----------
        source_dir : pathlib.Path
            Path to the source code directory
        project_dir : pathlib.Path
            Path to the root directory of the codebase being analysed
        package_name: str, optional. Default = None.
            The name of the codebase being analysed as it would be imported
        """
        for pyfile in source_dir.glob("**/*.py"):
            with open(pyfile, "r") as f:
                _pyfile_contents = f.readlines()

            self._version.review(pyfile, _pyfile_contents)
            del _pyfile_contents

        if (project_dir / "requirements.txt").exists():
            self._reqs = True

        self._setup.parse(project_dir, package_name)

    def __str__(self) -> str:
        reqs = str(self._version) + "\n"

        if self._reqs:
            reqs += "Install requirements with `pip install -r requirements.txt`.\n"

        reqs += str(self._setup) + "\n"

        if cli_entrypoints := self._setup.read_entrypoints():
            reqs += f"\n\n **Getting started**\n\n{cli_entrypoints}"

        return reqs


class DevOnboardingParser:
    def __init__(self) -> None:

        self._dev_reqs = False
        self._conda: str = ""
        self._pre_commit = False

        self._test_dir: Optional[str] = None
        self._test_libs: List[str] = []

    def parse(self, project_dir) -> None:
        """
        Search a project codebase for "Getting started" information useful to developers.

        Useful "Getting started" information includes the presence of
        separate developer requirements or setup instructions.

        Notes
        -----
        * Detects use of pre-commit
        * Detects presence of a conda environment
        * Detects presence of separate developer requirements
        * Infers testing environment
        """
        if (dev_reqs_file := (project_dir / "requirements-dev.txt")).exists():
            self._dev_reqs = True
            # TODO parse from API

        for _conda_filetype in ("yml", "yaml"):
            if (
                conda_file := (project_dir / f"environment.{_conda_filetype}")
            ).exists():
                self._parse_conda(conda_file)

        if (project_dir / ".pre-commit-config.yaml").exists():
            self._pre_commit = True

        self._parse_tests(project_dir)

    def _parse_tests(self, project_dir: Path) -> None:
        """
        Infer a test folder and framework.

        Detects "test" or "tests" folder as containing automated tests.
        Makes no distinction between types of tests.
        Infers if pytest or unittest testing frameworks are used.
        Some naive assumptions here.
        We should make the inferences more robust.
        """
        for _test_folder in ("tests", "test"):
            if (project_dir / _test_folder).is_dir():
                self._test_dir = _test_folder
                break

        if self._test_dir:
            for test_file in (project_dir / self._test_dir).glob("**/test_*.py"):
                with open(test_file, "r") as f:
                    test_contents = f.read()

                if "pytest" not in self._test_libs and (
                    "import pytest" in test_contents or "from pytest" in test_contents
                ):
                    self._test_libs.append("pytest")

                if "unittest" not in self._test_libs and (
                    "unittest.main" in test_contents or "TestCase):" in test_contents
                ):
                    self._test_libs.append("unittest")

    def _parse_conda(self, file: Path):
        # TODO
        self._conda = file.name

    def __str__(self) -> str:
        reqs = ""
        req_commands = []

        if self._dev_reqs:
            req_commands.append(
                "Install developer requirements with `pip install -r requirements-dev.txt`"
            )

        if self._conda:
            req_commands.append(
                f"To create an isolated development environment with anaconda, run `conda env create -f {self._conda}`"
            )

        if self._pre_commit:
            req_commands.append(
                "This project uses `pre-commit` to enforce code style on commit. Run `pre-commit install` in a terminal to setup"
            )

        if self._test_dir:
            test_info = f"Tests are present in `{self._test_dir}/`"

            if testing_framework := ", ".join(self._test_libs):
                test_info += f" (using {testing_framework})"

            req_commands.append(test_info)

        if req_commands:
            reqs += "* " + "\n* ".join(req_commands)

        return reqs + "\n\n"



from jaepeto.utils.config import parse_config
from jaepeto.utils.git import get_func_update, get_line_changes_for_file
from jaepeto.utils.parse_python import parse_all_type_hints
from jaepeto.utils.requests import post



"""
Write technical docs
"""
import os
import re
from datetime import datetime
from pathlib import Path
from textwrap import dedent
from typing import List, Optional, Tuple

from jaepeto.patterns import (
    CIParser,
    ContainerParser,
    describe_file_architecture,
    read_packages,
)
from jaepeto.structure import (
    DevOnboardingParser,
    FileAnalyser,
    OnboardingParser,
    get_script_entrypoints,
    group_imports_at_level,
    infer_important_functions,
)
from jaepeto.summaries import summarise_function
from jaepeto.utils import get_func_update, get_line_changes_for_file


def create_technical_doc(
    source_module_name: str,
    package_name: str,
    project_root: Path,
    technical_doc_file: Path,
    config,
    repo_info: str = None,
    project_title: str = "Project",
    project_summary: Optional[str] = None,
):
    if not technical_doc_file.exists():
        technical_doc_file.touch()

    with open(technical_doc_file, "r") as f:
        contents = f.readlines()

    contents = update_code_sections(contents)

    top_fixed_contents, bottom_fixed_contents = get_fixed_sections(contents)
    del contents

    # Add sections
    if top_fixed_contents == []:
        write_document_intro(technical_doc_file, project_title, project_summary)

    create_architecture_section(project_root, technical_doc_file)
    create_src_overview(
        source_module_name, package_name, project_root, technical_doc_file, config
    )
    create_developer_section(project_root, technical_doc_file)
    create_module_overview(source_module_name, project_root, technical_doc_file)

    # Re-add fixed sections
    with open(technical_doc_file, "r") as f:
        contents = f.readlines()

    contents = top_fixed_contents + contents + bottom_fixed_contents
    with open(technical_doc_file, "w") as f:
        f.writelines(contents)


def write_document_intro(technical_doc_file: Path, project_title: str, project_summary):
    with open(technical_doc_file, "w") as file:
        file.write(f"# {(project_title + ' Technical Documentation').strip()}")
        file.write(f"\n\n**Last updated:** {datetime.now().strftime('%Y-%m-%d')}\\")
        file.write("\n_Document generation aided by **Jaepeto**_")

        if project_summary:
            file.write(f"\n\n{project_summary}")

        file.write("\n\n* [Introduction](#introduction)")
        file.write("\n* [Code Overview](#code-overview)")

        file.write("\n\n## Introduction")
        file.write(
            f"""\n\nThis is a technical document detailing
        at a high-level
        what {project_title} does, how it operates,
        and how it is built.
        It is meant for consumption internally
        by technical developers
        and managers."""
        )

        file.write("\n\n### Scope")
        file.write(
            """\n\nThis document provides high-level summaries
        of the project codebase."""
        )

        file.write("\n\n### Context")
        file.write(
            """\n\nThe outline of this document was generated
        by **Jaepeto**.
        The documentation should be reviewed
        and detail provided manually
        before the documentation is released."""
        )


def create_architecture_section(project_root: Path, technical_doc_file: Path):
    arch_contents = ["\n\n## Project Architecture\n\n"]

    if packages := read_packages(project_root):
        arch_contents.append("The project has the following dependencies:\n")
        for _package in packages:
            arch_contents.append(f"\n* {_package}")
        arch_contents.append("\n\n")

    container_parser = ContainerParser()
    if container_description := container_parser.parse(project_root):
        container_arch = add_section_comments(
            "container", "arch", [container_description, "\n\n"]
        )
        arch_contents.extend(container_arch)

    arch_contents = add_section_comments("arch", "group", arch_contents)

    with open(technical_doc_file, "a") as file:
        file.writelines(arch_contents)


def create_developer_section(project_root: Path, technical_doc_file: Path) -> None:
    contents = ["\n\n## Developers"]

    # Requirements
    dev_onboarding = DevOnboardingParser()
    dev_onboarding.parse(project_root)

    if dev_setup := str(dev_onboarding):
        setup_info = add_section_comments("setup", "dev", [dev_setup, "\n\n"])
        contents.extend(setup_info)

    # CI
    ci_parser = CIParser()
    ci_parser.parse(project_root)

    if ci_description := str(ci_parser):
        ci_info = add_section_comments("ci", "dev", [ci_description, "\n\n"])
        contents.extend(ci_info)

    contents = add_section_comments("dev", "group", contents)

    with open(technical_doc_file, "a") as file:
        file.writelines(contents)


def create_src_overview(
    module_name: str,
    package_name: str,
    project_root: Path,
    technical_doc_file: Path,
    config,
) -> None:
    """
    Write high-level overview of project to technical doc.

    Minimum version.
    Entrypoints and entrypoint summaries.
    CI/CD processes used.
    Dockerfiles.
    Writes this information to technical document in project.

    Parameters
    ----------
    module_name : str
        The path from project root to the source code. In a project with a src directory, this would be `src/myproj`.
    package_name : str
        The name of the project package as it would get installed.
        In a project with a src directory `src/myproj`, this would probably be `myproj`.
    project_root : pathlib.Path
        Full path to project root
    technical_doc_file : pathlib.Path
        Path to technical document in project
    config : dict
        A fully-parsed config dict
    """
    source_code_dir = project_root / module_name
    contents = ["\n\n## Code Overview\n\n"]

    # Describe setup
    setup_parser = OnboardingParser()
    setup_parser.parse(source_code_dir, project_root, package_name)

    if setup_description := str(setup_parser):
        setup_contents = add_section_comments(
            "setup", "helloworld", [setup_description, "\n\n"]
        )
        contents.extend(setup_contents)
        del setup_contents

    # Important / __main__/__init__ functions
    entrypoint_contents = ["\n\n## Entrypoints\n\n"]
    entrypoints = infer_important_functions(package_name, source_code_dir)
    num_entrypoints = len(entrypoints)

    if 0 < num_entrypoints <= config["settings"]["max_num_entrypoints_threshold"]:
        entrypoint_contents.append(
            f"There are {num_entrypoints} source code objects in top-level `__main__`/`__init__`:\n\n"
        )
        for _entrypoint in entrypoints:
            entrypoint_func = _entrypoint.split(".")[-1]
            entrypoint_file = (os.sep).join(_entrypoint.split(".")[:-1]) + ".py"
            entrypoint_contents.append(
                f"### `{entrypoint_func}` from `{entrypoint_file}`\n"
            )

            _entrypoint_doc = summarise_function(
                entrypoint_func, project_root / entrypoint_file
            )
            if _entrypoint_doc:
                entrypoint_contents.append(f" \n```python\n{_entrypoint_doc}\n```\n")
            else:
                entrypoint_contents.append("\nObject cannot be summarised.\n")

        entrypoint_contents.append(
            "\nThese entrypoints are broken down into the following modules:\n\n"
        )
    elif num_entrypoints == 0:
        entrypoint_contents.append(
            f"There are 0 source code entrypoints in top-level `__main__`/`__init__` files.\n"
        )
    else:
        entrypoint_contents.append(
            f"""There are {num_entrypoints} source code objects
                        in top-level `__main__`/`__init__`.
                        They are broken down into the following modules:\n\n"""
        )

    entrypoint_groups = group_imports_at_level(entrypoints, level=2)

    for _entrypoint, _count in entrypoint_groups:
        entrypoint_contents.append(f"* `{_entrypoint}` has {_count} entrypoints\n")

    entrypoint_contents.append("\n")

    # Scripts with __name__ == __main__ and function calls
    script_entrypoints = get_script_entrypoints(source_code_dir, project_root)
    for _file, _entrypoints in script_entrypoints.items():
        if not _entrypoints:
            contents.append(f"`{_file}`` has a `__main__` entrypoint.\n\n")
        else:
            contents.append(f"`{_file}` has a `__main__` entrypoint, which calls:\n\n")
            for _entrypoint in _entrypoints:
                contents.append(f"* `{_entrypoint}`\n")
            contents.append("\n")

    contents.extend(
        add_section_comments("entrypoints", "helloworld", entrypoint_contents)
    )
    del entrypoint_contents

    # Classes used around the codebase
    class_type_hints = FileAnalyser(package_name, source_code_dir).parse()
    important_classes = []
    class_contents = []

    for class_name, class_uses in class_type_hints.items():
        if len(class_uses[1]) >= config["settings"]["num_class_uses_threshold"]:
            important_classes.append(
                f"`{class_name}` is a base class. It is used {len(class_uses[1])} times."
            ) if class_uses[0] is None else important_classes.append(
                f"`{class_name} inherits from {', '.join(class_uses[1])}"
            )

    if important_classes:
        class_contents.append(
            "The project has classes which are used in multiple functions:\n\n"
        )
        for _important_class in important_classes:
            class_contents.append(f"* {_important_class}\n")
        class_contents.append("\n")

        contents.extend(add_section_comments("classes", "helloworld", class_contents))
        del class_contents

    contents = add_section_comments("helloworld", "group", contents)

    with open(technical_doc_file, "a") as file:
        file.writelines(contents)


def create_module_overview(
    module_name: str, project_root: Path, technical_doc_file: Path
):
    """
    Create an overview of Python file structure in the project.
    """
    doc_contents, _ = _create_module_overview(
        project_root / module_name,
        project_root,
        [f"\n### **{project_root.stem}/**"],
        0,
    )

    if len(doc_contents) > 1:
        # if == 1, then only header was present but no python files. Therefore exclude.
        with open(technical_doc_file, "a") as file:
            file.writelines(doc_contents)


def _create_module_overview(module_path: Path, root: Path, doc_contents, n_calls: int):
    """
    Create an overview of Python file structure in the project.

    Include directories and subdirectories if they contain at least 1 python file.
    For each file, describe the rate of the change of the file,
    an inferred owner, and consequences of the file such as API or database calls.

    Parameters
    ----------
    module_path : pathlib.Path
        The path to the current directory being explored
    root : pathlib.Path
        The path to the top-level project directory.
        Used to infer local path to file (from top-level dir).
    doc_contents : list of str
        List containing contents above in the directory structure
        to be added to documentation
    n_calls : int
        DEPRECATED. Number of calls to API in this module overview section

    Returns
    -------
    list or str
        The contents to add to the documentation
    int
        DEPRECATED. The total number of API calls in this section
    """
    hash_level = min(6, len(doc_contents[-1].replace("\n", "").split()[0]) + 1)

    for subdir in module_path.glob("*"):
        if subdir.stem == "__pycache__":
            continue

        if subdir.is_dir():
            new_doc_contents, n_calls = _create_module_overview(
                subdir, root, [f"\n\n{'#' * hash_level} {subdir.stem}/"], n_calls
            )
            if len(new_doc_contents) > 1:
                doc_contents.extend(new_doc_contents)

    for py_file in module_path.glob("*.py"):
        if py_file.stem == "__init__":
            continue

        relative_file_path = str(py_file.relative_to(root))
        lines_added, lines_removed, most_common_author = get_line_changes_for_file(
            relative_file_path
        )

        if (lines_added + lines_removed) >= 20:
            doc_contents.append("\n\n")
            doc_contents.append(f"<!---Jaepeto-section-file: {relative_file_path}--->")
            doc_contents.append(f"\n\n{'#' * hash_level} {py_file.stem}.py")

            file_summary_str = dedent(
                f"""
                \n\nFile has {lines_added} lines added and {lines_removed} lines removed
                in the past 4 weeks."""
            )

            if most_common_author:
                file_summary_str += f" {most_common_author} is the inferred code owner."

            doc_contents.append(file_summary_str + "\n")

        if file_arch := describe_file_architecture(py_file):
            doc_contents.append(file_arch)

    return doc_contents, n_calls


def get_fixed_sections(doc_contents: List[str]) -> Tuple[List[str], List[str]]:
    """
    Get all of the top-fixed and bottom-fixed doc sections from the doc.

    Sections can be fixed to "top" or "bottom".
    Top-fixed sections have new content appended below them.
    Bottom-fixed sections have new content appended above them.

    Parameters
    ----------
    doc_contents : list of str
        The document

    Returns
    -------
    list of str
        Section contents which have been fixed to the top of the doc
    list of str
        Section contents which have been fixed to the bottom of the doc
    """
    return _get_fixed_section("top", doc_contents), _get_fixed_section(
        "bottom", doc_contents
    )


def _get_fixed_section(position: str, doc_contents: List[str]) -> List[str]:
    section = []
    section_count = 1

    while True:
        if section_content := get_doc_section(
            f"{position}{section_count}-start",
            f"{position}{section_count}-end",
            "fixed",
            doc_contents,
        ):
            section.extend(section_content)
            section_count += 1
        else:
            break

    return section


def has_section(name: str, section_type: str, doc_contents: List[str]) -> bool:
    comment = f"<!---Jaepeto-section-{section_type}: {name}-" + "{}--->\n"
    return (
        comment.format("start") in doc_contents
        and comment.format("end") in doc_contents
    )


def update_doc_code_sections(doc_file: Path) -> None:
    with open(doc_file, "r") as f:
        contents = f.readlines()

    contents = update_code_sections(contents)

    with open(doc_file, "w") as f:
        f.writelines(contents)


def update_code_sections(doc_contents: List[str]) -> List[str]:
    """
    Update every code section in a document to the latest API.

    Code sections are marked by a Jaepeto-section-code comment.
    Code sections do not need an -end comment.
    Code sections are terminated by open-closed ``` (code indicators).
    If another section starts before a code section is opened or closed,
    the block is not recognised as a code section.

    Parameters
    ----------
    doc_contents : list of str
        The document contents in which to search for code

    Returns
    -------
    list of str
        Document contents with code sections updated for latest commit.
    """
    new_code_blocks = []

    for idx, line in enumerate(doc_contents):
        if line.startswith("<!---Jaepeto-section-code:"):
            if new_code_info := update_code_section(idx, doc_contents):
                new_code_blocks.append(new_code_info)

    updated_doc_contents = doc_contents.copy()
    for start_idx, end_idx, new_code in new_code_blocks:
        updated_doc_contents = (
            updated_doc_contents[:start_idx]
            + [new_code]
            + updated_doc_contents[end_idx + 1 :]
        )

    return updated_doc_contents


def update_code_section(
    start_idx: int, doc_contents: List[str]
) -> Optional[Tuple[int, int, List[str]]]:
    """
    Identify and update a code snippet based on latest revision.

    To be updated, the comment attached to a code snippet
    must have the name of form {hash-start}-{file}-{line-start}-{line-end}.
    If opening or closing ``` are not detected, the code section is not updated.

    Parameters
    ----------
    start_idx : int
        The index of the beginning of a suspected code block to be updated
        within `doc_contents`
    doc_contents : list of str
        Documentation containing code blocks

    Returns
    -------
    None or (int, int, list of str)
        int = The start index of the OLD code block
        int = The end index of the OLD code block
        list of str = The NEW code block (inc. comment)
        Return None if code block not detected OR did not update.
    """
    comment = doc_contents[start_idx]
    if code_info_match := re.search(
        r" ([a-zA-Z0-9]{5,8})-(.+)-([0-9]{1,5})-([0-9]{1,5})", comment
    ):
        commit = code_info_match.group(1)
        code_file = code_info_match.group(2)
        code_start_line = int(code_info_match.group(3))
        code_end_line = int(code_info_match.group(4))

    object_name = None
    object_is_class = None
    code_started = False

    for idx, doc_line in enumerate(doc_contents[start_idx + 1 :]):
        doc_line = doc_line.strip()

        if doc_line.startswith("```"):
            if code_started and object_name:
                # Have reached the end of the code block
                # and have detected a trackable code object
                try:
                    if new_func_update := get_func_update(
                        code_file, object_name, commit
                    ):
                        latest_commit, _, new_file, new_code = new_func_update

                        new_code_selected_line = "\n".join(
                            new_code.split("\n")[: code_end_line - code_start_line + 1]
                        )
                        new_code_str = dedent(
                            f"""
<!---Jaepeto-section-code: {latest_commit}-{new_file}-{code_start_line}-{code_end_line}--->
```python
{new_code_selected_line}
```
"""
                        )
                        return start_idx, idx, new_code_str
                    else:
                        return None
                except:
                    return None
            else:
                code_started = True
        elif func_match := re.match(r"def (\w+)\(", doc_line):
            object_name = func_match.group(1)
            object_is_class = False
        elif class_match := re.match(r"class (\w+)[:\(]", doc_line):
            object_name = class_match.group(1)
            object_is_class = True

    return None


# --- Doc section management
def add_section_comments(
    section_name: str, section_type: str, contents: List[str]
) -> List[str]:
    """
    Bookend documentation contents with section-marking markdown comments.

    Prepends a `section-name`-start comment
    and append a `section-name`-end comment.
    """
    comment_base = f"<!---Jaepeto-section-{section_type}: {section_name}-" + "{}--->"
    return (
        ["\n", comment_base.format("start"), "\n"]
        + contents
        + ["\n", comment_base.format("end"), "\n"]
    )


def append_to_section(
    new_content: str, all_content: List[str], section_name: str, section_type: str
) -> List[str]:
    """
    Add documentation to a Jaepeto auto-generated section.

    Sections are detected by "Jaepeto-section" markdown comments.
    If a section comment with name "section_name"-end does not exist,
    the section is assumed to have ended when the next "Jaepeto-section"
    comment appears in the file.
    If the comment section is not in the file contents,
    returns file contents without the new contents added.

    Parameters
    ----------
    new_content : str
        The content to add to the documentation
    all_content : list of str
        The current documentation contents
    section_name : str
        The section name in the markdown comment to look for, in which to place the new content
    section_type : str
        The type of section to look out for. Should be "file"

    Returns
    -------
    list of str
        `all_content` with `new_content` added to the given section, if it exists
    """
    updated_content = all_content.copy()

    if not new_content.endswith("\n"):
        new_content += "\n"

    try:
        section_index_end = updated_content.index(
            f"<!---Jaepeto-section-{section_type}: {section_name}-end--->\n"
        )
        updated_content.insert(section_index_end, new_content)
        updated_content.insert(section_index_end + 1, "\n")
        return updated_content
    except ValueError:
        try:
            section_index = updated_content.index(
                f"<!---Jaepeto-section-{section_type}: {section_name}--->\n"
            )
        except ValueError:
            return updated_content
        else:
            for index in range(section_index + 1, len(updated_content)):
                if updated_content[index].startswith("<!---Jaepeto-section"):
                    updated_content.insert(index, new_content)
                    return updated_content

            # If got to here, no new section start, so just put new content at the end
            updated_content.append(new_content)
            return updated_content


def get_doc_section(
    section_start_name: str,
    section_end_name: str,
    section_type: str,
    doc_contents: List[str],
) -> List[str]:
    try:
        section_start_index = doc_contents.index(
            f"<!---Jaepeto-section-{section_type}: {section_start_name}--->\n"
        )
        section_end_index = doc_contents.index(
            f"<!---Jaepeto-section-{section_type}: {section_end_name}--->\n"
        )
    except ValueError:
        return []
    else:
        return doc_contents[section_start_index : section_end_index + 1]


def replace_doc_section(
    section_start_name: str,
    section_end_name: str,
    section_type: str,
    new_contents: List[str],
    doc_contents: List[str],
) -> List[str]:
    try:
        section_start_index = doc_contents.index(
            f"<!---Jaepeto-section-{section_type}: {section_start_name}--->\n"
        )
        section_end_index = doc_contents.index(
            f"<!---Jaepeto-section-{section_type}: {section_end_name}--->\n"
        )
    except ValueError:
        return doc_contents
    else:
        new_doc_contents = doc_contents.copy()
        new_doc_contents[section_start_index : section_end_index + 1] = new_contents
        return new_doc_contents
